/****************************************************************************/
/* Mar/2007                                                                 */
/* Peter Boyle.                                                             */
/*                                                                          */
/* bfm.h                                                                   */
/* C++ header file DWF & Wilson CG inverter with SIMD support               */
/*                                                                          */
/****************************************************************************/

#ifndef INCLUDED_DWF_PAB_H
#define INCLUDED_DWF_PAB_H

#include <math.h>
#include <string.h>
#include <bfm_options.h>
#include <bagel_int.h>
#include <stdlib.h>
#include <list>
#ifdef BFM_QDP_BINDING
#include <qdp.h>
#define CHROMA_3
#endif

typedef double QDPdouble ;
enum { mem_fast, mem_sendbuf, mem_slow };

class bfmarg {

public:
   int node_latt[4];  /*Node subvolume*/
   int local_comm[4]; /*are comms periodic/trivial in this axis      */
   int ncoor[4];      /*location of node in 4d node grid*/
   int Ls;            /*Fifth dimension - set to 1 for wilson/clover */
   double M5;         /*Five-D mass- dont care for wilson/clover */
   double mass;       /*fermion mass*/
   double Csw;        /*Clover coefficient - unimplemented as yet*/
   int max_iter;      // CG control
   double residual;   // CG control
   int precon_5d;     // Boolean selection of 4d or 5d checkerboarding in DWF solves.
                      // Set to 0 for Wilson solves.
   int reproduce;


   bfmarg(){
     Csw=0;
     reproduce = 0;
     max_iter=5000;
     residual = 1.0e-7;
     precon_5d = 0;
     ncoor[0] = 0;
     ncoor[1] = 0;
     ncoor[2] = 0;
     ncoor[3] = 0;
   };

};



/* DWF matrix */

// Internally the code works only with a logical/data-parallel subnodes.
// Interaction between logical subnodes only happens during comms
// This class operates on "internal" layout vectors
// The CG routine will support import/export to more standard orderings.
  /* Constant definitions */
  enum {
    ND  = 4,
    SPINOR_SIZE = 24,
    HALF_SPINOR_SIZE = 12,
    GAUGE_SIZE = 72,
    Nmu = 4,
    Ncb = 2,
    NMinusPlus = 2,
    Minus=0,
    Plus =1
  };

typedef void * Fermion_t;

template <class Float>
class bfmbase : public bfmarg {

 protected:

  int nsimd;

 public:


  int ncb;
  int node_cbvol;
  int cbLs;
  int simd_latt[ND];
  int simd_grid[ND];
  integer simd_cbvol;
  int simd_nbound[ND];
  int simd_allbound;
  int base_parity;

  Float *u;

  // Reproducability checking
  enum { ReproIdle, ReproRecord , ReproCheck };
  void repro_control(int state) ;
  void repro(double d) ;


  std::list<double> repro_log;
  int repro_state;
  // End hack
/*
 * Build up enough infrastructure to write a generic solver
 * with abstracted internal structure.
 *
 * With good luck and a following wind, I should be able to cover 
 * Wilson/Clover/TwistedMass/DWF in one go.
 *
 * I will explicitly implement CG below for the congenitally lazy.
 * 
 * I would love to write an equivalent "C" output for SSE3.
 *
 */


 public:

/******************************************************************************************************
 * Public Methods that are common to all implementations are in the base class
 ******************************************************************************************************/


  /**************************************************************************
   * This is the QDP visible glue
   **************************************************************************
   */
#ifdef BFM_QDP_BINDING
  virtual void   importGauge(multi1d<LatticeColorMatrix> &u);
  void           importFermion  (LatticeFermion &psi, Fermion_t handle,int cb,int s=0);
  void           exportFermion  (LatticeFermion &psi, Fermion_t handle,int cb,int s=0);

  // for DWF
  void      importFermion  (multi1d<LatticeFermion> &psi, Fermion_t handle,int cb);
  void      exportFermion  (multi1d<LatticeFermion> &psi, Fermion_t handle,int cb);

  // Internal interfaces
  int CG4d(LatticePropagator &sol, 
	   LatticePropagator &src,
	   multi1d<LatticeColorMatrix> &U,
	   std::basic_string<char> output_stem);
#endif

  /**************************************************************************
   * Generic interface for use in creating external solvers
   **************************************************************************
   */
  Fermion_t allocFermion   (int mem_type=mem_slow); 
  void      freeFermion    (Fermion_t handle);

  void impexFermion  (QDPdouble *psi, Fermion_t handle, int import, int cb,int s) ;
  void      importFermion  (QDPdouble *psi, Fermion_t handle,int cb);
  void      exportFermion  (QDPdouble *psi, Fermion_t handle,int cb);
  virtual void importGauge(QDPdouble *gauge, int dir) ;

  virtual int doubleStored(void);

  /*Linalg*/
  double    norm(Fermion_t psi);
  void      scale(Fermion_t psi,double a);
  void      axpy(Fermion_t result, Fermion_t x, Fermion_t y,double a);
  void      axpby(Fermion_t result, Fermion_t x, Fermion_t y,double a,double b);
  double    axpy_norm(Fermion_t result, Fermion_t x, Fermion_t y,double a);
  double    axpby_norm(Fermion_t result, Fermion_t x, Fermion_t y,double a,double b);
  void      chiralproj_axpby(Fermion_t y, Fermion_t x, double a, int sign, int s_slices);

  ////////////////////////////////////////////////////////////////////////////////////
  // My solver of choice
  ////////////////////////////////////////////////////////////////////////////////////

  int CGNE(Fermion_t sol_guess[2], Fermion_t source[2]);
  int CGNE_prec(Fermion_t sol_guess, Fermion_t source);

  void dump(Fermion_t psi, char *filename, double scale,int cb);

  double Mprec(Fermion_t chi, 
	      Fermion_t psi, 
	      Fermion_t tmp, 
	      int dag,int donrm=0) ;

  void Munprec(Fermion_t chi[2], 
	        Fermion_t psi[2], 
	        Fermion_t tmp,
		int dag) ;

  /*Mooee depends on 4d/5d precon versus */
  /*bagel_idx and bagel_idx5d depend on 4d/5d precon versus */
  /*dslash depends (weakly) on 4d/5d precon versus */

  void Mooee5dprec(Fermion_t psi, 
	                   Fermion_t chi, 
	                   int dag);
  void MooeeInv5dprec(Fermion_t psi, 
	            Fermion_t chi, 
	            int dag);
  void Mooee4dprec(Fermion_t psi, 
	                   Fermion_t chi, 
	                   int dag);
  void MooeeInv4dprec(Fermion_t psi, 
	            Fermion_t chi, 
	            int dag);


  void Mooee(Fermion_t chi, 
		     Fermion_t psi, 
		     int dag) ;


  void MooeeInv(Fermion_t chi, 
			Fermion_t psi, 
			int dag) ;

  integer chroma_idx(int site[4],int reim,int i, int i_size) ;
  int mprecFlopsPerSite ( void ) ;

/***************************************************************************************************************************
 * Public Methods that are specific to certain implementations are virtual
 ***************************************************************************************************************************/

  virtual void init(bfmarg &arg) = 0; 
  virtual void end (void) = 0;

  virtual integer bagel_idx   (int site[4],int reim,int i, int i_size,int half_cb) = 0;
  virtual integer bagel_idx5d (int site[4],int s,int reim,int i, int i_size,int half_cb) =0 ;
  virtual integer bagel_idx5d (int site,int s,int i_size) =0;

  virtual void dslash(Fermion_t chi, 
		      Fermion_t psi, 
		      int cb,
		      int dag,int dperp=1) = 0;


  virtual void Dperp(Fermion_t psi, 
		     Fermion_t chi, 
		     int cb,
		     int dag)=0;

  virtual int simd(void) = 0;
  virtual void pointers_init(void) = 0;
  virtual void pointers_end(void) = 0; 
  virtual void comm_init(void) = 0;
  virtual void comm_end(void) = 0;
  virtual void comm(int result_cb, Fermion_t psi,int dag) = 0; 
  virtual void comm_start(int result_cb, Fermion_t psi,int dag) = 0; 
  virtual void comm_complete(int result_cb, Fermion_t psi) = 0; 
  virtual void comm_gsum(double &val);
  virtual bool isBoss();

};

// Other prototypes
#include <bfm_vmx.h>
void *bfm_alloc(size_t size,int mem_type = mem_slow);
void  bfm_free (void *ptr);


#ifdef BFM_BAGEL1_QMP
#include <bfmbagel1.h>
#include <bfmcommqmpbagel1.h>

typedef commQMPbagel1<double> bfm;
typedef commQMPbagel1<double> bfm_dp;
typedef commQMPbagel1<float>  bfm_sp;

template <class Float>
class bfm_internal : public commQMPbagel1<Float> { 
};

#endif

#ifdef BFM_BAGEL1_FAKE
#include <bfmbagel1.h>
#include <bfmcommfakebagel1.h>

typedef commfakebagel1<double> bfm;
typedef commfakebagel1<double> bfm_dp;
typedef commfakebagel1<float>  bfm_sp;
template <class Float>
class bfm_internal : public commfakebagel1<Float> { 
};

#endif

#ifdef BFM_BAGEL2_QMP
#include <bfmbagel2.h>
#include <bfmcommqmpbagel2.h>

typedef commQMPbagel2<double> bfm;
typedef commQMPbagel2<double> bfm_dp;
typedef commQMPbagel2<float>  bfm_sp;
template <class Float>
class bfm_internal : public commQMPbagel2<Float> { 
};
#endif

#ifdef BFM_BAGEL2_FAKE
#include <bfmbagel2.h>
#include <bfmcommfakebagel2.h>

typedef commfakebagel2<double> bfm;
typedef commfakebagel2<double> bfm_dp;
typedef commfakebagel2<float>  bfm_sp;

template <class Float>
class bfm_internal : public commfakebagel2<Float> { 
};
#endif

#endif
